from typing import List

from twisted.internet import defer, protocol
from twisted.python import reflect

from txjsonrpc_ng import jsonrpclib


class BaseSubhandler:
    """
    Sub-handlers for prefixed methods (e.g., system.listMethods)
    can be added with putSubHandler. By default, prefixes are
    separated with a '.'. Override self.separator to change this.
    """
    separator = '.'

    def __init__(self):
        self.subHandlers = {}

    def putSubHandler(self, prefix, handler):
        self.subHandlers[prefix] = handler

    def getSubHandler(self, prefix):
        return self.subHandlers.get(prefix, None)

    def getSubHandlerPrefixes(self):
        return self.subHandlers.keys()

    def _getFunction(self, functionPath):
        """
        Given a string, return a function, or raise jsonrpclib.NoSuchFunction.

        This returned function will be called, and should return the result
        of the call, a Deferred, or a Fault instance.

        Override in subclasses if you want your own policy. The default
        policy is that given functionPath 'foo', return the method at
        self.jsonrpc_foo, i.e. getattr(self, "jsonrpc_" + functionPath).
        If functionPath contains self.separator, the sub-handler for
        the initial prefix is used to search for the remaining path.
        """
        if functionPath.find(self.separator) != -1:
            prefix, functionPath = functionPath.split(self.separator, 1)
            handler = self.getSubHandler(prefix)
            if handler is None:
                raise jsonrpclib.NoSuchFunction(jsonrpclib.METHOD_NOT_FOUND,
                                                "no such sub-handler %s" % prefix)
            return handler._getFunction(functionPath)
        if functionPath == '__dir__':
            return self._listFunctions
        f = getattr(self, "jsonrpc_%s" % functionPath, None)
        if not f:
            raise jsonrpclib.NoSuchFunction(jsonrpclib.METHOD_NOT_FOUND,
                                            "function %s not found" % functionPath)
        elif not callable(f):
            raise jsonrpclib.NoSuchFunction(jsonrpclib.METHOD_NOT_CALLABLE,
                                            "function %s not callable" % functionPath)
        else:
            return f

    def _listFunctions(self):
        """
        Return a list of the names of all jsonrpc methods.
        """
        return reflect.prefixedMethodNames(self.__class__, 'jsonrpc_')


class BaseQueryFactory(protocol.ClientFactory):
    deferred = None
    protocol = None

    # XXX add an "id" parameter
    id = 0

    def __init__(self, method, version=jsonrpclib.VERSION_PRE1, *args):
        # XXX pass the "id" parameter here
        self.version = version
        self.id = self.id + 1
        self.payload = self._buildVersionedPayload(method, args)
        self.deferred = defer.Deferred()

    def _buildVersionedPayload(self, *args):
        if self.version == jsonrpclib.VERSION_PRE1:
            return jsonrpclib._preV1Request(*args)
        elif self.version == jsonrpclib.VERSION_1:
            return jsonrpclib._v1Request(*args, id=self.id)
        elif self.version == jsonrpclib.VERSION_2:
            return jsonrpclib._v2Request(*args, id=self.id)

    def parseResponse(self, contents):
        if not self.deferred:
            return
        try:
            # Convert the response from JSON-RPC to python.
            result = jsonrpclib.loads(contents)
            if self.version != jsonrpclib.VERSION_PRE1:
                result = result["result"]
            elif isinstance(result, list):
                result = result[0]
        except Exception as error:
            self.deferred.errback(error)
            self.deferred = None
        else:
            self.deferred.callback(result)
            self.deferred = None

    def clientConnectionFailed(self, _, reason):
        if self.deferred is not None:
            self.deferred.errback(reason)
            self.deferred = None

    clientConnectionLost = clientConnectionFailed

    def badStatus(self, status, message):
        self.deferred.errback(ValueError(status, message))
        self.deferred = None


class BaseProxy:
    """
    A Proxy base class for making remote JSON-RPC calls.
    """

    def __init__(self, version=jsonrpclib.VERSION_PRE1, factoryClass=None):
        self.version = version
        self.factoryClass = factoryClass

    def _getVersion(self, keywords):
        version = keywords.get("version")
        if version is None:
            version = self.version
        return version

    def _getFactoryClass(self, keywords):
        factoryClass = keywords.get("factoryClass")
        if not factoryClass:
            factoryClass = self.factoryClass
        return factoryClass


class Introspection(BaseSubhandler):
    """
    Implement the JSON-RPC Introspection API.

    By default, the methodHelp method returns the 'help' method attribute,
    if it exists, otherwise the __doc__ method attribute, if it exists,
    otherwise the empty string.

    To enable the methodSignature method, add a 'signature' method attribute
    containing a list of lists. See methodSignature's documentation for the
    format. Note the type strings should be JSON-RPC types, not Python types.
    """

    def __init__(self, parent):
        """
        Implement Introspection support for an JSONRPC server.

        @param parent: the JSONRPC server to add Introspection support to.
        """
        BaseSubhandler.__init__(self)
        self._jsonrpc_parent = parent

    def jsonrpc_listMethods(self) -> List[str]:
        """
        Return a list of the method names implemented by this server.
        """
        functions = []
        todo = [(self._jsonrpc_parent, '')]
        while todo:
            obj, prefix = todo.pop(0)
            functions.extend([prefix + name for name in obj._listFunctions()])
            todo.extend([(obj.getSubHandler(name),
                          prefix + name + obj.separator)
                         for name in obj.getSubHandlerPrefixes()])
        functions.sort()
        return functions

    def jsonrpc_methodHelp(self, method: str) -> str:
        """
        Return a documentation string describing the use of the given method.
        """
        method = self._jsonrpc_parent._getFunction(method)
        return (getattr(method, 'help', None)
                or getattr(method, '__doc__', None) or '').strip()

    def jsonrpc_methodSignature(self, method: str) -> str:
        """
        Return a list of type signatures.

        Each type signature is a list of the form [rtype, type1, type2, ...]
        where rtype is the return type and typeN is the type of the Nth
        argument. If no signature information is available, the empty
        string is returned.
        """
        method = self._jsonrpc_parent._getFunction(method)
        return getattr(method, 'signature', None) or ''


def addIntrospection(jsonrpc):
    """
    Add Introspection support to an JSONRPC server.

    @param jsonrpc: The jsonrpc server to add Introspection support to.
    """
    # jsonrpc.putSubHandler('system', Introspection, ('protocol',))
    jsonrpc.putSubHandler('system', Introspection(jsonrpc))



display_name = "txJSON-RPC"
library_name = "txjsonrpc"
version = "0.4"
author = "Duncan McGreggor"
author_email = "oubiwann@adytum.us"
license = "BSD, GPL"
url = "http://launchpad.net/%s" % library_name
description = "Code for creatig Twisted JSON-RPC servers and clients."



"""
Requires simplejson; can be downloaded from
http://cheeseshop.python.org/pypi/simplejson
"""
try:
    import xmlrpclib
except ImportError:

    import xmlrpc.client as xmlrpclib

from datetime import datetime

import json

# From xmlrpclib.
SERVER_ERROR = xmlrpclib.SERVER_ERROR
NOT_WELLFORMED_ERROR = xmlrpclib.NOT_WELLFORMED_ERROR
UNSUPPORTED_ENCODING = xmlrpclib.UNSUPPORTED_ENCODING
INVALID_ENCODING_CHAR = xmlrpclib.INVALID_ENCODING_CHAR
INVALID_JSONRPC = xmlrpclib.INVALID_XMLRPC
METHOD_NOT_FOUND = xmlrpclib.METHOD_NOT_FOUND
INVALID_METHOD_PARAMS = xmlrpclib.INVALID_METHOD_PARAMS
INTERNAL_ERROR = xmlrpclib.INTERNAL_ERROR

# Custom errors.
METHOD_NOT_CALLABLE = -32604

# Version constants.

VERSION_PRE1 = 0
VERSION_1 = 1
VERSION_2 = 2

Fault = xmlrpclib.Fault


class NoSuchFunction(Fault):
    """
    There is no function by the given name.
    """


class JSONRPCEncoder(json.JSONEncoder):
    """
    Provide custom serializers for JSON-RPC.
    """

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y%m%dT%H:%M:%S")
        raise TypeError("%r is not JSON serializable" % (obj,))


def dumps(obj, **kwargs):
    try:
        version = kwargs.pop("version")
    except KeyError:
        version = VERSION_PRE1
    try:
        id = kwargs.pop("id")
    except KeyError:
        id = None

    if isinstance(obj, Exception):
        result = None
        if version != VERSION_2:
            error = {'fault': obj.__class__.__name__,
                     'faultCode': obj.faultCode,
                     'faultString': obj.faultString}
        else:
            error = {'message': obj.faultString,
                     'code': obj.faultCode,
                     'data': ''}
    else:
        result = obj
        error = None
    if version == VERSION_PRE1:
        if result:
            obj = result
        else:
            obj = error
    elif version == VERSION_1:
        obj = {"result": result, "error": error, "id": id}
    elif version == VERSION_2:
        if error:
            obj = {"jsonrpc": "2.0", "error": error, "id": id}
        else:
            obj = {"jsonrpc": "2.0", "result": result, "id": id}
    else:
        obj = {"result": result, "error": error, "id": id}
    return json.dumps(obj, cls=JSONRPCEncoder, **kwargs)


def loads(string, **kws):
    unmarshalled = json.loads(string, **kws)
    # XXX there's going to need to be some version-conditional code here...
    # for versions greater than VERSION_PRE1, we'll have to check for the
    # "error" key, not the "fault" key... and then raise if "fault" is not
    # None.
    if (isinstance(unmarshalled, dict) and "fault" in unmarshalled):
        raise Fault(unmarshalled['faultCode'], unmarshalled['faultString'])
    if (isinstance(unmarshalled, dict) and "error" in unmarshalled):
        if "jsonrpc" in unmarshalled and unmarshalled["jsonrpc"] == "2.0":
            raise Fault(unmarshalled["error"]['code'], unmarshalled["error"]['message'])
        if unmarshalled['error']:
            raise Fault(unmarshalled["error"]['faultCode'], unmarshalled["error"]['faultString'])
    return unmarshalled


class SimpleParser(object):
    buffer = ''

    def feed(self, data):
        self.buffer += data

    def close(self):
        self.data = loads(self.buffer)


class SimpleUnmarshaller(object):

    def getmethodname(self):
        return self.parser.data.get("method")

    def getid(self):
        return self.parser.data.get("id")

    def close(self):
        if isinstance(self.parser.data, dict):
            return self.parser.data.get("params")
        return self.parser.data


def getparser():
    parser = SimpleParser()
    marshaller = SimpleUnmarshaller()
    marshaller.parser = parser
    return parser, marshaller


class Transport(xmlrpclib.Transport):
    """
    Handles an HTTP transaction to an XML-RPC server.
    """
    user_agent = "jsonrpclib.py (by txJSON-RPC)"

    def getparser(self):
        """
        Get Parser and unmarshaller.
        """
        return getparser()


def _preV1Request(method="", params=[], *args):
    return dumps({"method": method, "params": params})


def _v1Request(method="", params=[], id="", *args):
    return dumps(
        {"method": method, "params": params, "id": id})


def _v1Notification(method="", params=[], *args):
    return _v1Request(method=method, params=params, id=None)


def _v2Request(method="", params=[], id="", *args):
    return dumps({
        "jsonrpc": "2.0", "method": method, "params": params, "id": id})


def _v2Notification(method="", params=[], *args):
    return _v2Request(method=method, params=params, id=None)


class ServerProxy(xmlrpclib.ServerProxy):
    """
    XXX add missing docstring
    """

    def __init__(self, uri, transport=Transport(), version=VERSION_PRE1, *args,
                 **kwds):
        xmlrpclib.ServerProxy.__init__(self, uri, transport, *args, **kwds)
        self.version = version

    def __request(self, *args):
        """
        Call a method on the remote server.

        XXX Is there any way to indicate that we'd want a notification request
        instead of a regular request?
        """
        request = self._getVersionedRequest(*args)
        # XXX do a check here for id; if null, skip the response
        # XXX in order to do this effectively, we might have to change the
        # request functions to objects, so that we can get at an id attribute
        response = self.__transport.request(
            self.__host,
            self.__handler,
            request,
            verbose=self.__verbose
        )
        if len(response) == 1:
            response = response[0]
        return response

    def _getVersionedRequest(self, *args):
        if self.version == VERSION_PRE1:
            return _preV1Request(*args)
        elif self.version == VERSION_1:
            return _v1Request(*args)
        elif self.version == VERSION_2:
            return _v2Request(*args)



import os


legalReSTFiles = [
    "README.rst",
    "TODO",
    "DEPENDENCIES",
    ]


def setup(*args, **kwds):
    """
    Compatibility wrapper.
    """
    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup
    return setup(*args, **kwds)


def findPackages(library_name):
    """
    Compatibility wrapper.

    Taken from storm setup.py.
    """
    try:
        from setuptools import find_packages
        return find_packages()
    except ImportError:
        pass
    packages = []
    for directory, subdirectories, files in os.walk(library_name):
        if "__init__.py" in files:
            packages.append(directory.replace(os.sep, "."))
    return packages


def hasDocutils():
    """
    Check to see if docutils is installed.
    """
    try:
        import docutils
        return docutils
    except ImportError:
        return False


def _validateReST(text):
    """
    Make sure that the given ReST text is valid.

    Taken from Zope Corp's zc.twist setup.py.
    """
    import docutils.utils
    import docutils.parsers.rst
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

    doc = docutils.utils.new_document("validator")
    # our desired settings
    doc.reporter.halt_level = 5
    doc.reporter.report_level = 1
    stream = doc.reporter.stream = StringIO()
    # docutils buglets (?)
    doc.settings.tab_width = 2
    doc.settings.pep_references = doc.settings.rfc_references = False
    doc.settings.trim_footnote_reference_space = None
    # and we're off...
    parser = docutils.parsers.rst.Parser()
    parser.parse(text, doc)
    return stream.getvalue()


def validateReST(text):
    """
    A wrapper that ensafens the validation for pythons that are not embiggened
    with docutils.
    """
    if hasDocutils():
        return _validateReST(text)
    print(" *** No docutils; can't validate ReST.")
    return ""


def catReST(*args, **kwds):
    """
    Concatenate the contents of one or more ReST files.

    Taken from Zope Corp's zc.twist setup.py.
    """
    # note: distutils explicitly disallows unicode for setup values :-/
    # http://docs.python.org/dist/meta-data.html
    tmp = []
    for arg in args:
        if arg in legalReSTFiles or arg.endswith(".txt"):
            f = open(os.path.join(*arg.split("/")))
            tmp.append(f.read())
            f.close()
            tmp.append("\n\n")
        else:
            print("Warning: '%s' not a legal ReST filename.")
            tmp.append(arg)
    if len(tmp) == 1:
        res = tmp[0]
    else:
        res = "\n\n".join(tmp)
    out = kwds.get("out")
    stop_on_errors = kwds.get("stop_on_errors")
    if out is True:
        out = "CHECK_THIS_BEFORE_UPLOAD.txt"
    if out:
        f = open(out, "w")
        f.write(res)
        f.close()
        report = validateReST(res)
        if report:
            print(report)
            if stop_on_errors:
                print("""ReST validation error

See the following:
  http://docutils.sourceforge.net/docs/user/rst/cheatsheet.txt
  http://docutils.sourceforge.net/docs/user/rst/quickstart.html
""")
                raise ValueError("ReST validation error")
    return res



import abc
import gzip
import io
import time
from collections.abc import Callable
from typing import Any, Optional

from twisted.web.http import Request

from .data import CacheableResult


class Renderer(metaclass=abc.ABCMeta):

    def __init__(self, id: str, version: int, request: Request):
        self.id = id
        self.version = version
        self.request = request

    @abc.abstractmethod
    def render(self, string_renderer: Callable[[Any, str, int], str]) -> None:
        pass

    def handle_compression(self, response_string: str, cached_response: Optional[bytes],
                           cache_updater: Optional[Callable[[bytes], None]]) -> None:
        compression = self.request.getHeader('Accept-encoding')
        original_size = len(response_string)
        if compression == "gzip" and original_size >= 1000:
            if cached_response is not None:
                response_binary = cached_response
            else:
                start_time = time.time()
                out_file = io.BytesIO()
                with gzip.GzipFile(mode='wb', fileobj=out_file) as in_file:
                    in_file.write(response_string.encode())
                response_binary = out_file.getvalue()

                compressed_size = len(response_binary)
                elapsed_time = time.time() - start_time
                break_even = (original_size - compressed_size) / elapsed_time / 1024 / 1024
                print("renderer: compress data {} -> {} ({:.1f} %) in {:.2f} ms (break even at {:.1f} MB/s)".format(original_size,
                                                                                                          compressed_size,
                                                                                                          compressed_size * 100 / original_size,
                                                                                                          elapsed_time * 1000,
                                                                                                          break_even))
                if cache_updater is not None:
                    cache_updater(response_binary)

            self.request.setHeader("content-encoding", "gzip")
        else:
            response_binary = response_string.encode()

        self.request.setHeader(b"content-length", str(len(response_binary)))
        self.request.write(response_binary)


class DefaultRenderer(Renderer):

    def __init__(self, result: Any, id: str, version: int, request: Request):
        super().__init__(id, version, request)
        self.result = result

    def render(self, string_renderer: Callable[[Any, str, int], str]) -> None:
        result_string = string_renderer(self.result, self.id, self.version)

        self.handle_compression(result_string, None, None)


class CacheableResultRenderer(Renderer):

    def __init__(self, result: CacheableResult, id: str, version: int, request: Request):
        super().__init__(id, version, request)
        self.result = result

    def render(self, call: Callable[[Any, str, int], str]) -> None:
        if self.result.string_value is not None:
            string_value = self.result.string_value
        else:
            string_value = call(self.result.value, self.id, self.version)
            self.result.string_value = string_value

        def update_value(compressed_value: bytes) -> None:
            self.result.compressed_value = compressed_value

        self.handle_compression(
            string_value,
            self.result.compressed_value,
            update_value
        )


def renderer_factory(result, id, version, request: Request):
    if isinstance(result, CacheableResult):
        return CacheableResultRenderer(result, id, version, request)
    else:
        return DefaultRenderer(result, id, version, request)



# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
# See LICENSE for details.
"""
A generic resource for publishing objects via JSON-RPC.

API Stability: semi-stable

Maintainer: U{Duncan McGreggor <mailto:oubiwann@adytum.us>}
"""
from twisted.internet import defer, protocol, reactor
from twisted.protocols import basic
from twisted.python import log

from txjsonrpc_ng import jsonrpclib
from txjsonrpc_ng.jsonrpc import (
    BaseProxy, BaseQueryFactory, BaseSubhandler, Introspection)


class JSONRPC(basic.NetstringReceiver, BaseSubhandler):
    """
    A protocol that implements JSON-RPC.

    Methods published can return JSON-RPC serializable results, Faults,
    Binary, Boolean, DateTime, Deferreds, or Handler instances.

    By default methods beginning with 'jsonrpc_' are published.
    """
    # Error codes for Twisted, if they conflict with yours then
    # modify them at runtime.
    NOT_FOUND = 8001
    FAILURE = 8002

    separator = '.'
    closed = 0


    def __init__(self, version=jsonrpclib.VERSION_2):
        BaseSubhandler.__init__(self)
        self.version = version

    def __call__(self):
        return self

    def connectionMade(self):
        self.MAX_LENGTH = self.factory.maxLength

    def stringReceived(self, line):
        parser, unmarshaller = jsonrpclib.getparser()
        deferred = defer.maybeDeferred(parser.feed, line.decode())

        req, req_id = self._cbDispatch(parser, unmarshaller)
        deferred.addCallback(lambda x: req)
        deferred.addErrback(self._ebRender, req_id = req_id)
        deferred.addCallback(self._cbRender, req_id = req_id)
        return deferred

    def _cbDispatch(self, parser, unmarshaller):
        parser.close()
        args, functionPath, req_id  = unmarshaller.close(), unmarshaller.getmethodname(), unmarshaller.getid()
        function = self._getFunction(functionPath)
        return defer.maybeDeferred(function, *args), req_id

    def _cbRender(self, result, req_id):
        if self.version == jsonrpclib.VERSION_PRE1 and not isinstance(result, jsonrpclib.Fault):
            result = (result,)
        try:
            s = jsonrpclib.dumps(result, id=req_id, version=self.version)
        except:
            f = jsonrpclib.Fault(self.FAILURE, "can't serialize output")
            s = jsonrpclib.dumps(f, id=req_id, version=self.version)
        return self.sendString(s.encode())

    def _ebRender(self, failure, req_id):
        if isinstance(failure.value, jsonrpclib.Fault):
            return failure.value
        log.err(failure)
        return jsonrpclib.Fault(self.FAILURE, "error")


class QueryProtocol(basic.NetstringReceiver):

    def connectionMade(self):
        self.data = ''
        msg = self.factory.payload
        packet = '%d:%s,' % (len(msg), msg)
        self.transport.write(packet.encode())

    def stringReceived(self, string):
        self.factory.data = string.decode()
        self.transport.loseConnection()


class QueryFactory(BaseQueryFactory):

    protocol = QueryProtocol
    data = ''

    def clientConnectionLost(self, _, reason):
        self.parseResponse(self.data)


class Proxy(BaseProxy):
    """
    A Proxy for making remote JSON-RPC calls.

    Pass the URL of the remote JSON-RPC server to the constructor.

    Use proxy.callRemote('foobar', *args) to call remote method
    'foobar' with *args.
    """

    def __init__(self, host, port, version=jsonrpclib.VERSION_PRE1,
                 factoryClass=QueryFactory):
        """
        @type host: C{str}
        @param host: The host to which method calls are made.

        @type port: C{integer}
        @param port: The host's port to which method calls are made.

        @type version: C{int}
        @param version: The version indicates which JSON-RPC spec to support.
        The available choices are jsonrpclib.VERSION*. The default is to use
        the version of the spec that txJSON-RPC was originally released with,
        pre-Version 1.0.

        @type factoryClass: C{object}
        @param factoryClass: The factoryClass should be a subclass of
        QueryFactory (class, not instance) that will be used instead of
        QueryFactory.
        """
        BaseProxy.__init__(self, version, factoryClass)
        self.host = host
        self.port = port

    def callRemote(self, method, *args, **kwargs):
        version = self._getVersion(kwargs)
        factoryClass = self._getFactoryClass(kwargs)
        factory = factoryClass(method, version, *args)
        reactor.connectTCP(self.host, self.port, factory)
        return factory.deferred


class RPCFactory(protocol.ServerFactory):

    protocol = None

    def __init__(self, rpcClass, maxLength=1024):
        self.maxLength = maxLength
        self.protocol = rpcClass
        self.subHandlers = {}

    def buildProtocol(self, addr):
        p = protocol.ServerFactory.buildProtocol(self, addr)
        for key, val in self.subHandlers.items():
            klass, args, kws = val
            if args and args[0] == 'protocol':
                p.putSubHandler(key, klass(p))
            else:
                p.putSubHandler(key, klass(*args, **kws))
        return p

    def putSubHandler(self, name, klass, args=(), kws={}):
        self.subHandlers[name] = (klass, args, kws)

    def addIntrospection(self):
        self.putSubHandler('system', Introspection, ('protocol',))


__all__ = ["JSONRPC", "Proxy", "RPCFactory"]



"""
Utility functions for testing.
"""
import os
import unittest
import doctest


def importModule(name):
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def fileIsTest(path, skipFiles=[]):
    if not os.path.isfile(path):
        return False
    filename = os.path.basename(path)
    if filename in skipFiles:
        return False
    if filename.startswith('test') and filename.endswith('.py'):
        return True


def find(start, func, skip=[]):
    for item in [os.path.join(start, x) for x in os.listdir(start)]:
        if func(item, skip):
            yield item
        if os.path.isdir(item):
            for subItem in find(item, func, skip):
                yield subItem


def findTests(startDir, skipFiles=[]):
    return find(startDir, fileIsTest, skipFiles)


def buildDoctestSuite(modules):
    suite = unittest.TestSuite()
    for modname in modules:
        mod = importModule(modname)
        suite.addTest(doctest.DocTestSuite(mod))
    return suite


def buildUnittestSuites(paths=[], skip=[]):
    """
    paths: a list of directories to search
    skip: a list of file names to skip
    """
    suites = []
    loader = unittest.TestLoader()
    for startDir in paths:
        for testFile in findTests(startDir, skip):
            modBase = os.path.splitext(testFile)[0]
            name = modBase.replace(os.path.sep, '.')
            # import the testFile as a module
            mod = importModule(name)
            # iterate through module objects, checking for TestCases
            for objName in dir(mod):
                if not objName.endswith('TestCase'):
                    continue
                obj = getattr(mod, objName)
                if not issubclass(obj, unittest.TestCase):
                    continue
                # create a suite from any test cases
                suite = loader.loadTestsFromTestCase(obj)
                # append to suites list
                suites.append(suite)
    return suites



from twisted import web
from twisted.cred.portal import IRealm, Portal
from twisted.web import guard
from zope.interface import implementer


@implementer(IRealm)
class HTTPAuthRealm(object):

    def __init__(self, resource):
        self.resource = resource

    def logout(self):
        pass

    def requestAvatar(self, avatarId, mind, *interfaces):
        if web.resource.IResource in interfaces:
            return web.resource.IResource, self.resource, self.logout
        raise NotImplementedError()


def wrapResource(resource, checkers, credFactories=[], realmName=""):
    defaultCredFactory = guard.BasicCredentialFactory(realmName)
    credFactories.insert(0, defaultCredFactory)
    realm = HTTPAuthRealm(resource)
    portal = Portal(realm, checkers)
    return guard.HTTPAuthSessionWrapper(portal, credFactories)



# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
# See LICENSE for details.
"""
A generic resource for publishing objects via JSON-RPC.

Requires simplejson; can be downloaded from
http://cheeseshop.python.org/pypi/simplejson

API Stability: unstable

Maintainer: U{Duncan McGreggor<mailto:oubiwann@adytum.us>}
"""

import codecs
import io

from .render import renderer_factory

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

try:
    import xmlrpclib
except ImportError:
    import xmlrpc.client as xmlrpclib
import gzip

from twisted.web import resource, server
from twisted.internet import defer, reactor
from twisted.python import log, context
from twisted.web import http

from txjsonrpc_ng import jsonrpclib
from txjsonrpc_ng.jsonrpc import BaseProxy, BaseQueryFactory, BaseSubhandler

# Useful so people don't need to import xmlrpclib directly.
Fault = xmlrpclib.Fault
Binary = xmlrpclib.Binary
Boolean = xmlrpclib.Boolean
DateTime = xmlrpclib.DateTime


def with_request(method):
    """
    Decorator to enable the request to be passed as the first argument.
    """
    method.with_request = True
    return method


def requires_auth(method):
    """
    Decorator to enable authentication before resolving the method.
    """
    method.requires_auth = True
    return method


class NoSuchFunction(Fault):
    """
    There is no function by the given name.
    """


class Unauthorized(jsonrpclib.Fault):
    def __init__(self, message):
        Fault.__init__(self, 4000, message)


class Handler:
    """
    Handle a JSON-RPC request and store the state for a request in progress.

    Override the run() method and return result using self.result,
    a Deferred.

    We require this class since we're not using threads, so we can't
    encapsulate state in a running function if we're going  to have
    to wait for results.

    For example, lets say we want to authenticate against twisted.cred,
    run a LDAP query and then pass its result to a database query, all
    as a result of a single JSON-RPC command. We'd use a Handler instance
    to store the state of the running command.
    """

    def __init__(self, resource, *args):
        # the JSON-RPC resource we are connected to
        self.resource = resource
        self.result = defer.Deferred()
        self.run(*args)

    def run(self, *args):
        # event driven equivalent of 'raise UnimplementedError'
        self.result.errback(
            NotImplementedError("Implement run() in subclasses"))


class JSONRPC(resource.Resource, BaseSubhandler):
    """
    A resource that implements JSON-RPC.

    Methods published can return JSON-RPC serializable results, Faults,
    Binary, Boolean, DateTime, Deferreds, or Handler instances.

    By default methods beginning with 'jsonrpc_' are published.
    """

    # Error codes for Twisted, if they conflict with yours then
    # modify them at runtime.
    NOT_FOUND = 8001
    FAILURE = 8002

    isLeaf = 1
    except_map = {}
    auth_token = "Auth-Token"

    def __init__(self):
        resource.Resource.__init__(self)
        BaseSubhandler.__init__(self)

    def render(self, request):
        request.content.seek(0, 0)
        # Unmarshal the JSON-RPC data.
        content = request.content.read().decode()
        if not content and request.method == 'GET' and 'request' in request.args:
            content = request.args['request'][0]
        self.callback = request.args['callback'][0] if 'callback' in request.args else None
        self.is_jsonp = True if self.callback else False
        if not self.is_jsonp:
            request.setHeader("content-type", "application/json")
        else:
            request.setHeader("content-type", "text/javascript")
        parsed = jsonrpclib.loads(content)
        functionPath = parsed.get("method")
        params = parsed.get('params', {})
        args, kwargs = [], {}
        if params.__class__ == list:
            args = params
        else:
            kwargs = params
        id = parsed.get('id')
        token = None
        if request.requestHeaders.hasHeader(self.auth_token):
            token = request.requestHeaders.getRawHeaders(self.auth_token)[0]
        version = parsed.get('jsonrpc')
        if version:
            version = int(float(version))
        elif id and not version:
            version = jsonrpclib.VERSION_1
        else:
            version = jsonrpclib.VERSION_PRE1
        # XXX this all needs to be re-worked to support logic for multiple
        # versions...
        try:
            function = self._getFunction(functionPath)
            d = None
            if hasattr(function, 'requires_auth'):
                d = defer.maybeDeferred(self.auth, token, functionPath)
        except jsonrpclib.Fault as f:
            self._cbRender(f, request, id, version)
        else:
            if hasattr(function, 'with_request'):
                args = [request] + args

            if d:
                d.addCallback(context.call, function, *args, **kwargs)
            else:
                d = defer.maybeDeferred(function, *args, **kwargs)
            d.addErrback(self._ebRender, id)
            d.addCallback(self._cbRender, request, id, version)

            def _responseFailed(err, call):
                call.cancel()

            request.notifyFinish().addErrback(_responseFailed, d)
        return server.NOT_DONE_YET

    def _cbRender(self, result, request, id, version):
        if isinstance(result, Handler):
            result = result.result

        if result is not None:
            renderer = renderer_factory(result, id, version, request)
            renderer.render(self._render_text)

        request.finish()
        return result

    def _render_text(self, result, id, version) -> str:
        if version == jsonrpclib.VERSION_PRE1:
            if not isinstance(result, jsonrpclib.Fault):
                result = (result,)
        try:
            s = jsonrpclib.dumps(result, id=id, version=version) if not self.is_jsonp else "%s(%s)" % (
                self.callback, jsonrpclib.dumps(result, id=id, version=version))
        except:
            f = jsonrpclib.Fault(self.FAILURE, "can't serialize output")
            s = jsonrpclib.dumps(f, id=id, version=version) if not self.is_jsonp else "%s(%s)" % (
                self.callback, jsonrpclib.dumps(f, id=id, version=version))
        return s

    def _map_exception(self, exception):
        return self.except_map.get(exception, self.FAILURE)

    def _ebRender(self, failure, id):
        if isinstance(failure.value, jsonrpclib.Fault):
            return failure.value
        log.err(failure)
        message = failure.value.message if hasattr(failure.value, 'message') else repr(failure.value)
        code = self._map_exception(type(failure.value))
        return jsonrpclib.Fault(code, message)

    def auth(self, token, func):
        return True


class QueryProtocol(http.HTTPClient):

    def __init__(self):
        self.response_headers = {}

    def connectionMade(self):
        self.sendCommand(b'POST', self.factory.path.encode())
        self.sendHeader(b'User-Agent', b'Twisted/JSONRPClib')
        self.sendHeader(b'Host', self.factory.host.encode())
        self.sendHeader(b'Content-type', b'application/json')
        if self.factory.compress:
            self.sendHeader(b'Accept-encoding', b'gzip')
        self.sendHeader(b'Content-length', str(len(self.factory.payload)))
        if self.factory.user:
            auth = '%s:%s' % (self.factory.user, self.factory.password)
            auth = codecs.encode(auth.encode(), 'base64')
            self.sendHeader(b'Authorization', b'Basic ' + auth)
        self.endHeaders()
        self.transport.write(self.factory.payload.encode())

    def handleStatus(self, version, status, message):
        status = status.decode()
        if status != '200':
            self.factory.badStatus(status, message.decode())

    def handleHeader(self, key, val):
        self.response_headers[key.decode().lower()] = val.decode()

    def handleResponse(self, contents):
        if self.response_headers.get("content-encoding") == "gzip":
            compressed_file = io.BytesIO(contents)
            with gzip.GzipFile(mode='rb', fileobj=compressed_file) as in_file:
                contents = in_file.read()
            compressed_file.close()
        self.factory.parseResponse(contents.decode())


class QueryFactory(BaseQueryFactory):
    deferred = None
    protocol = QueryProtocol

    def __init__(self, path, host, method, user=None, password=None,
                 version=jsonrpclib.VERSION_PRE1, compress=False, *args):
        BaseQueryFactory.__init__(self, method, version, *args)
        self.path, self.host = path, host
        self.user, self.password = user, password
        self.compress = compress


class Proxy(BaseProxy):
    """
    A Proxy for making remote JSON-RPC calls.

    Pass the URL of the remote JSON-RPC server to the constructor.

    Use proxy.callRemote('foobar', *args) to call remote method
    'foobar' with *args.
    """

    def __init__(self, url, user=None, password=None,
                 version=jsonrpclib.VERSION_PRE1, compress=False, factoryClass=QueryFactory, ssl_ctx_factory=None):
        """
        @type url: C{str}
        @param url: The URL to which to post method calls.  Calls will be made
        over SSL if the scheme is HTTPS.  If netloc contains username or
        password information, these will be used to authenticate, as long as
        the C{user} and C{password} arguments are not specified.

        @type user: C{str} or None
        @param user: The username with which to authenticate with the server
        when making calls.  If specified, overrides any username information
        embedded in C{url}.  If not specified, a value may be taken from C{url}
        if present.

        @type password: C{str} or None
        @param password: The password with which to authenticate with the
        server when making calls.  If specified, overrides any password
        information embedded in C{url}.  If not specified, a value may be taken
        from C{url} if present.

        @type version: C{int}
        @param version: The version indicates which JSON-RPC spec to support.
        The available choices are jsonrpclib.VERSION*. The default is to use
        the version of the spec that txJSON-RPC was originally released with,
        pre-Version 1.0.

        @type ssl_ctx_factory: C{twisted.internet.ssl.ClientContextFactory} or None
        @param ssl_ctx_factory: SSL client context factory class to use instead
        of default twisted.internet.ssl.ClientContextFactory.
        """
        BaseProxy.__init__(self, version, factoryClass)
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        netlocParts = netloc.split('@')
        if len(netlocParts) == 2:
            userpass = netlocParts.pop(0).split(':')
            self.user = userpass.pop(0)
            try:
                self.password = userpass.pop(0)
            except:
                self.password = None
        else:
            self.user = self.password = None
        hostport = netlocParts[0].split(':')
        self.host = hostport.pop(0)
        try:
            self.port = int(hostport.pop(0))
        except:
            self.port = None
        self.path = path
        if self.path in ['', None]:
            self.path = '/'
        self.secure = (scheme == 'https')
        if user is not None:
            self.user = user
        if password is not None:
            self.password = password
        self.compress = compress
        self.ssl_ctx_factory = ssl_ctx_factory

    def callRemote(self, method, *args, **kwargs):
        version = self._getVersion(kwargs)
        # XXX generate unique id and pass it as a parameter
        factoryClass = self._getFactoryClass(kwargs)
        factory = factoryClass(self.path, self.host, method, self.user,
                               self.password, version, self.compress, *args)
        if self.secure:
            from twisted.internet import ssl
            if self.ssl_ctx_factory is None:
                self.ssl_ctx_factory = ssl.ClientContextFactory
            reactor.connectSSL(self.host, self.port or 443,
                               factory, self.ssl_ctx_factory())
        else:
            reactor.connectTCP(self.host, self.port or 80, factory)
        return factory.deferred


__all__ = ["JSONRPC", "Handler", "Proxy"]



from dataclasses import dataclass
from typing import Union, Dict, List, Optional


@dataclass
class CacheableResult:
    value: Union[Dict, List]
    string_value: Optional[str] = None
    compressed_value: Optional[bytes] = None


